# Copyright 2022 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import torch
import torch.nn as nn
from typing import List, Union
from model_compression_toolkit.gptq.common.gptq_config import GradientPTQConfig
from model_compression_toolkit.gptq.pytorch.quantizer.gptq_quantizer import BaseWeightQuantizer
from model_compression_toolkit.core.pytorch.utils import to_torch_tensor
from model_compression_toolkit.gptq.pytorch.quantizer.quant_utils import ste_round, ste_clip
from model_compression_toolkit.gptq.common.gptq_constants import AUXVAR
from model_compression_toolkit.core.common.quantization.node_quantization_config import NodeWeightsQuantizationConfig
from model_compression_toolkit.core.common.constants import THRESHOLD


class STEWeightQuantizer(BaseWeightQuantizer):
    """
    Class that implements a quantizer with trainable parameters to be used for GPTQ training.
    """

    def __init__(self,
                 weights_quantization_cfg: NodeWeightsQuantizationConfig,
                 gptq_config: GradientPTQConfig,
                 weight: torch.nn.Parameter):
        """
        Construct a Pytorch model that utilize a fake weight quantizer of STE (Straight Through Estimator) for symmetric quantizer.
        Args:
            weights_quantization_cfg: Configuration of weight quantization
            gptq_config: GradientPTQConfig object with parameters about the tuning process.
            weight: weight for auxiliary tensor creation.
        """
        super().__init__()

        self.signed = True
        self.num_bits = weights_quantization_cfg.weights_n_bits
        self.min_int = -int(self.signed) * (2 ** (self.num_bits - int(self.signed)))
        self.max_int = (2 ** (self.num_bits - int(self.signed))) - 1
        self.weight_shape = weight.shape
        self.threshold_values = weights_quantization_cfg.weights_quantization_params.get(THRESHOLD)
        self.delta_tensor = self.threshold_values / (2 ** (self.num_bits-int(self.signed)))
        self.max_delta_change = gptq_config.lsb_change_per_bit_width.get(self.num_bits)

        # Set trainable tensors
        self.set_trainable_params()

        # Create tensors
        self.delta_tensor = to_torch_tensor(self.delta_tensor)
        self.max_tensor_change = self.delta_tensor * self.max_delta_change

    def set_trainable_params(self):
        """
        A function to set a list of trainable parameters of the quantizer for GPTQ retraining
        """
        self.aux_tensor = nn.Parameter(to_torch_tensor(torch.zeros(self.weight_shape)), requires_grad=True)
        self.trainable_params.update({AUXVAR: self.aux_tensor})

    def get_aux_variable(self) -> torch.Tensor:
        """
        Returns auxiliary trainable variables
        """
        return self.trainable_params.get(AUXVAR)

    def get_quantization_variable(self) -> Union[torch.Tensor, List]:
        """
        Returns quantization trainable variables
        """
        return []

    def get_weight_quantization_params(self) -> dict:
        """
        Returns weight quantization dictionary params
        """
        return {THRESHOLD: self.threshold_values}

    def forward(self, w: nn.Parameter, training: bool = True) -> nn.Parameter:
        """
        Weight fake quantizer
        Args:
            w: weights to quantize.
            training: whether in training mode or not
        Returns:
            quantized weights
        """
        v0 = ste_clip(self.aux_tensor, min_val=-self.max_tensor_change, max_val=self.max_tensor_change)
        v1 = v0 / self.delta_tensor
        w0 = torch.round(w / self.delta_tensor).detach()
        w1 = w0 + v1
        w2 = ste_round(w1)
        w3 = ste_clip(w2, min_val=self.min_int, max_val=self.max_int)
        w_q = self.delta_tensor * w3
        return w_q




# Copyright 2022 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================


from abc import abstractmethod
from typing import Any, Callable

from model_compression_toolkit.core.common import Logger


class Exporter:
    """
    Base class to define API for an Exporter class that exports and saves models.
    At initiation, it gets a model to export. This model must be an exportable model.
    Each exporter needs to implement a method called 'export' which export the model
    (convert an exportable model to a final model to run on the target platform),
    and a method named 'save_model' to allow the saving and serialization of the
    exported model.
    """

    def __init__(self, model: Any, is_layer_exportable_fn: Callable):
        """

        Args:
            model: Model to export.
            is_layer_exportable_fn: Callable to check whether a layer can be exported or not.

        """
        self.model = model
        self.is_layer_exportable_fn = is_layer_exportable_fn

    @abstractmethod
    def export(self) -> Any:
        """

        Returns: Converted model to export.

        """
        Logger.critical(f'Exporter {self.__class__} have to implement export method')

    @abstractmethod
    def save_model(self, save_model_path: str):
        """
        Save exported model in a given path.
        Args:
            save_model_path: Path to save the exported model.

        Returns:
            None.
        """
        Logger.critical(f'Exporter {self.__class__} have to implement save_model method')



# Copyright 2022 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from enum import Enum
from functools import partial
from typing import List

import numpy as np

from model_compression_toolkit import FrameworkInfo
from model_compression_toolkit.core.common import Graph, BaseNode
from model_compression_toolkit.core.common.constants import BITS_TO_BYTES, FLOAT_BITWIDTH
from model_compression_toolkit.core.common.framework_implementation import FrameworkImplementation
from model_compression_toolkit.core.common.graph.edge import EDGE_SINK_INDEX
from model_compression_toolkit.core.common.graph.virtual_activation_weights_node import VirtualActivationWeightsNode, \
    VirtualSplitWeightsNode, VirtualSplitActivationNode
from model_compression_toolkit.core.common.logger import Logger


def weights_size_kpi(mp_cfg: List[int],
                     graph: Graph,
                     fw_info: FrameworkInfo,
                     fw_impl: FrameworkImplementation) -> np.ndarray:
    """
    Computes a KPIs vector with the respective weights' memory size for the given weight configurable node,
    according to the given mixed-precision configuration.
    If an empty configuration is given, then computes KPI vector for non-configurable nodes.

    Args:
        mp_cfg: A mixed-precision configuration (list of candidates index for each configurable node)
        graph: Graph object.
        fw_info: FrameworkInfo object about the specific framework (e.g., attributes of different layers' weights to quantize).
        fw_impl: FrameworkImplementation object with specific framework methods implementation (not used in this method).

    Returns: A vector of node's weights memory sizes.
    Note that the vector is not necessarily of the same length as the given config.

    """
    weights_memory = []
    mp_nodes = graph.get_configurable_sorted_nodes_names()
    weights_mp_nodes = [n.name for n in graph.get_sorted_weights_configurable_nodes()]

    if len(mp_cfg) == 0:
        # Computing non-configurable nodes KPI
        for n in graph.nodes:
            if n.name not in mp_nodes and n.has_weights_quantization_enabled_candidate():
                if n.name not in weights_mp_nodes \
                        and n.has_weights_quantization_enabled_candidate():
                    node_nbits = n.candidates_quantization_cfg[0].weights_quantization_cfg.weights_n_bits
                    node_weights_memory_in_bytes = _compute_node_weights_memory(n, node_nbits, fw_info)
                    weights_memory.append(node_weights_memory_in_bytes)
                elif not n.reuse:
                    Logger.warning(f"Non-configurable nodes should have a single quantization configuration candidate,"
                                   f"but node {n.name} has {len(n.candidates_quantization_cfg)} candidates. "
                                   f"The node's weights memory is not considered for the weights KPI computation.")
    else:
        # Go over configurable all nodes that should be taken into consideration when computing the weights KPI.
        for n in graph.get_sorted_weights_configurable_nodes():
            node_idx = mp_nodes.index(n.name)
            node_qc = n.candidates_quantization_cfg[mp_cfg[node_idx]]
            node_nbits = node_qc.weights_quantization_cfg.weights_n_bits

            node_weights_memory_in_bytes = _compute_node_weights_memory(n, node_nbits, fw_info)

            weights_memory.append(node_weights_memory_in_bytes)

    return np.array(weights_memory)


def activation_output_size_kpi(mp_cfg: List[int],
                               graph: Graph,
                               fw_info: FrameworkInfo,
                               fw_impl: FrameworkImplementation) -> np.ndarray:
    """
    Computes a KPIs vector with the respective output memory size for each activation configurable node,
    according to the given mixed-precision configuration.
    If an empty configuration is given, then computes KPI vector for non-configurable nodes.

    Args:
        mp_cfg: A mixed-precision configuration (list of candidates index for each configurable node)
        graph: Graph object.
        fw_info: FrameworkInfo object about the specific framework (e.g., attributes of different layers' weights to quantize)
            (not used in this method).
        fw_impl: FrameworkImplementation object with specific framework methods implementation(not used in this method).

    Returns: A vector of node's activation memory sizes.
    Note that the vector is not necessarily of the same length as the given config.

    """
    activation_memory = []
    mp_nodes = graph.get_configurable_sorted_nodes_names()
    activation_mp_nodes = [n.name for n in graph.get_sorted_activation_configurable_nodes()]

    if len(mp_cfg) == 0:
        # Computing non-configurable nodes KPI
        for n in graph.nodes:
            if n.name not in activation_mp_nodes \
                    and n.has_activation_quantization_enabled_candidate():
                if n.is_all_activation_candidates_equal():
                    node_nbits = n.candidates_quantization_cfg[0].activation_quantization_cfg.activation_n_bits
                    node_activation_memory_in_bytes = _compute_node_activation_memory(n, node_nbits)
                    activation_memory.append(node_activation_memory_in_bytes)
                else:
                    Logger.warning(f"Non-configurable nodes should have a single quantization configuration candidate,"  # pragma: no cover
                                   f"but node {n.name} has {len(n.candidates_quantization_cfg)} candidates. "
                                   f"The node's activation memory is not considered for the weights KPI computation.")
    else:
        # Go over all nodes that should be taken into consideration when computing the weights KPI.
        for n in graph.get_sorted_activation_configurable_nodes():
            node_idx = mp_nodes.index(n.name)
            node_qc = n.candidates_quantization_cfg[mp_cfg[node_idx]]
            node_nbits = node_qc.activation_quantization_cfg.activation_n_bits

            node_activation_memory_in_bytes = _compute_node_activation_memory(n, node_nbits)

            activation_memory.append(node_activation_memory_in_bytes)

    return np.array(activation_memory)


def total_weights_activation_kpi(mp_cfg: List[int],
                                 graph: Graph,
                                 fw_info: FrameworkInfo,
                                 fw_impl: FrameworkImplementation) -> np.ndarray:
    """
    Computes KPIs tensor with the respective weights size and output memory size for each activation configurable node,
    according to the given mixed-precision configuration.
    If an empty configuration is given, then computes KPI vector for non-configurable nodes.

    Args:
        mp_cfg: A mixed-precision configuration (list of candidates index for each configurable node)
        graph: Graph object.
        fw_info: FrameworkInfo object about the specific framework (e.g., attributes of different layers' weights to quantize)
            (not used in this method).
        fw_impl: FrameworkImplementation object with specific framework methods implementation(not used in this method).

    Returns: A 2D tensor of nodes' weights memory sizes and activation output memory size.
    Note that the vector is not necessarily of the same length as the given config.

    """
    weights_activation_memory = []
    mp_nodes = graph.get_configurable_sorted_nodes_names()
    weights_mp_nodes = [n.name for n in graph.get_sorted_weights_configurable_nodes()]
    activation_mp_nodes = [n.name for n in graph.get_sorted_activation_configurable_nodes()]

    if len(mp_cfg) == 0:
        # Computing non-configurable nodes KPI
        for n in graph.nodes:

            non_configurable = False
            node_weights_memory_in_bytes, node_activation_memory_in_bytes = 0, 0

            # Non-configurable Weights
            if n.name not in mp_nodes:
                if n.name not in mp_nodes and n.has_weights_quantization_enabled_candidate():
                    if n.name not in weights_mp_nodes \
                            and n.has_weights_quantization_enabled_candidate():
                        node_nbits = n.candidates_quantization_cfg[0].weights_quantization_cfg.weights_n_bits
                        node_weights_memory_in_bytes = _compute_node_weights_memory(n, node_nbits, fw_info)
                        non_configurable = True

                    elif not n.reuse:
                        Logger.warning(
                            f"Non-configurable nodes should have a single quantization configuration candidate,"
                            f"but node {n.name} has {len(n.candidates_quantization_cfg)} candidates. "
                            f"The node's weights memory is not considered for the weights KPI computation.")

            # Non-configurable Activation
            if n.name not in activation_mp_nodes \
                    and n.has_activation_quantization_enabled_candidate():
                if n.is_all_activation_candidates_equal():
                    node_nbits = n.candidates_quantization_cfg[0].activation_quantization_cfg.activation_n_bits
                    node_activation_memory_in_bytes = _compute_node_activation_memory(n, node_nbits)
                    non_configurable = True

                else:
                    Logger.warning(f"Non-configurable nodes should have a single quantization configuration candidate,"  # pragma: no cover
                                   f"but node {n.name} has {len(n.candidates_quantization_cfg)} candidates. "
                                   f"The node's activation memory is not considered for the weights KPI computation.")

            if non_configurable:
                weights_activation_memory.append(
                    np.array([node_weights_memory_in_bytes, node_activation_memory_in_bytes]))
    else:
        # Go over all nodes that should be taken into consideration when computing the weights or
        # activation KPI (all configurable nodes).
        for node_idx, n in enumerate(graph.get_configurable_sorted_nodes()):
            node_qc = n.candidates_quantization_cfg[mp_cfg[node_idx]]
            node_weights_nbits = node_qc.weights_quantization_cfg.weights_n_bits
            node_activation_nbits = node_qc.activation_quantization_cfg.activation_n_bits

            # Compute node's weights memory (if no weights to quantize then set to 0)
            node_weights_memory_in_bytes = 0
            if n.is_weights_quantization_enabled() and not n.is_all_weights_candidates_equal():
                node_weights_memory_in_bytes = _compute_node_weights_memory(n, node_weights_nbits, fw_info)

            # Compute node's activation memory (if node's activation are not being quantized then set to 0)
            node_activation_memory_in_bytes = 0
            if n.is_activation_quantization_enabled() and not n.is_all_activation_candidates_equal():
                node_activation_memory_in_bytes = _compute_node_activation_memory(n, node_activation_nbits)

            weights_activation_memory.append(np.array([node_weights_memory_in_bytes, node_activation_memory_in_bytes]))

    return np.array(weights_activation_memory)


def bops_kpi(mp_cfg: List[int],
             graph: Graph,
             fw_info: FrameworkInfo,
             fw_impl: FrameworkImplementation,
             set_constraints: bool = True) -> np.ndarray:
    """
    Computes a KPIs vector with the respective bit-operations (BOPS) count for each configurable node,
    according to the given mixed-precision configuration of a virtual graph with composed nodes.

    Args:
        mp_cfg: A mixed-precision configuration (list of candidates index for each configurable node)
        graph: Graph object.
        fw_info: FrameworkInfo object about the specific framework (e.g., attributes of different layers' weights to quantize).
        fw_impl: FrameworkImplementation object with specific framework methods implementation.
        set_constraints: A flag for utilizing the method for KPI computation of a
            given config not for LP formalization purposes.

    Returns: A vector of node's BOPS count.
    Note that the vector is not necessarily of the same length as the given config.

    """

    if not set_constraints:
        return _bops_kpi(mp_cfg,
                         graph,
                         fw_info,
                         fw_impl)

    # BOPs KPI method considers non-configurable nodes, therefore, it doesn't need separate implementation
    # for non-configurable nodes for setting a constraint (no need for separate implementation for len(mp_cfg) = 0).

    virtual_bops_nodes = [n for n in graph.get_topo_sorted_nodes() if isinstance(n, VirtualActivationWeightsNode)]

    mp_nodes = graph.get_configurable_sorted_nodes_names()
    bops = [n.get_bops_count(fw_impl, fw_info, candidate_idx=_get_node_cfg_idx(n, mp_cfg, mp_nodes)) for n in virtual_bops_nodes]

    return np.array(bops)


def _bops_kpi(mp_cfg: List[int],
              graph: Graph,
              fw_info: FrameworkInfo,
              fw_impl: FrameworkImplementation) -> np.ndarray:
    """
    Computes a KPIs vector with the respective bit-operations (BOPS) count for each configurable node,
    according to the given mixed-precision configuration of an original graph.

    Args:
        mp_cfg: A mixed-precision configuration (list of candidates index for each configurable node)
        graph: Graph object.
        fw_info: FrameworkInfo object about the specific framework (e.g., attributes of different layers' weights to quantize).
        fw_impl: FrameworkImplementation object with specific framework methods implementation.

    Returns: A vector of node's BOPS count.

    """

    mp_nodes = graph.get_configurable_sorted_nodes_names()

    # Go over all nodes that should be taken into consideration when computing the BOPS KPI.
    bops = []
    for n in graph.get_topo_sorted_nodes():
        if n.has_weights_to_quantize(fw_info):
            # If node doesn't have weights then its MAC count is 0, and we shouldn't consider it in the BOPS count.
            incoming_edges = graph.incoming_edges(n, sort_by_attr=EDGE_SINK_INDEX)
            if len(incoming_edges) != 1:
                Logger.critical(f"Can't compute BOPS metric for node {n.name} with multiple inputs.")  # pragma: no cover

            input_activation_node = incoming_edges[0].source_node
            if len(graph.out_edges(input_activation_node)) > 1:
                # In the case where the activation node has multiple outgoing edges
                # we don't consider this edge in the BOPS KPI calculation
                continue

            input_activation_node_cfg = input_activation_node.candidates_quantization_cfg[_get_node_cfg_idx(input_activation_node, mp_cfg, mp_nodes)]

            node_mac = fw_impl.get_node_mac_operations(n, fw_info)

            node_qc = n.candidates_quantization_cfg[_get_node_cfg_idx(n, mp_cfg, mp_nodes)]
            node_weights_nbits = node_qc.weights_quantization_cfg.weights_n_bits if \
                node_qc.weights_quantization_cfg.enable_weights_quantization else FLOAT_BITWIDTH
            input_activation_nbits = input_activation_node_cfg.activation_quantization_cfg.activation_n_bits if \
                input_activation_node_cfg.activation_quantization_cfg.enable_activation_quantization else FLOAT_BITWIDTH

            node_bops = node_weights_nbits * input_activation_nbits * node_mac
            bops.append(node_bops)

    return np.array(bops)


def _get_node_cfg_idx(node: BaseNode, mp_cfg: List[int], sorted_configurable_nodes_names: List[str]) -> int:
    """
    Returns the index of a node's quantization configuration candidate according to the given
    mixed-precision configuration. If the node is not configurable, then it must have a single configuration,
    therefore, the index 0 is returned.

    Args:
        node: A node to get its candidate configuration index.
        mp_cfg: A mixed-precision configuration (list of candidates index for each configurable node)
        sorted_configurable_nodes_names: A list of configurable nodes names.

    Returns: An index (integer) of a node's quantization configuration candidate.
    """

    if node.name in sorted_configurable_nodes_names:
        node_idx = sorted_configurable_nodes_names.index(node.name)
        return mp_cfg[node_idx]
    else:
        assert len(node.candidates_quantization_cfg) > 0, \
            "Any node should have at least one candidate configuration."
        return 0


def _get_origin_weights_node(n: BaseNode) -> BaseNode:
    """
    In case we run a KPI computation on a virtual graph,
    this method is used to retrieve the original node out of a virtual weights node,

    Args:
        n: A possibly virtual node.

    Returns: A node from the original (non-virtual) graph which the given node represents.

    """

    if isinstance(n, VirtualActivationWeightsNode):
        return n.original_weights_node
    if isinstance(n, VirtualSplitWeightsNode):
        return n.origin_node

    return n


def _get_origin_activation_node(n: BaseNode) -> BaseNode:
    """
    In case we run a KPI computation on a virtual graph,
    this method is used to retrieve the original node out of a virtual activation node,

    Args:
        n: A possibly virtual node.

    Returns: A node from the original (non-virtual) graph which the given node represents.

    """

    if isinstance(n, VirtualActivationWeightsNode):
        return n.original_activation_node
    if isinstance(n, VirtualSplitActivationNode):
        return n.origin_node

    return n


def _compute_node_weights_memory(n: BaseNode, node_nbits: int, fw_info: FrameworkInfo) -> float:
    """
    Computes the weights' memory of the given node.

    Args:
        n: A node to compute its weights' memory.
        node_nbits: A bit-width in which the node's weights should be quantized.
        fw_info: FrameworkInfo object about the specific framework.

    Returns: The total memory of the node's weights when quantized to the given bit-width.

    """

    origin_node = _get_origin_weights_node(n)

    node_num_weights_params = 0
    for attr in fw_info.get_kernel_op_attributes(origin_node.type):
        if attr is not None:
            node_num_weights_params += origin_node.get_weights_by_keys(attr).flatten().shape[0]

    return node_num_weights_params * node_nbits / BITS_TO_BYTES


def _compute_node_activation_memory(n: BaseNode, node_nbits: int) -> float:
    """
    Computes the activation tensor memory of the given node.

    Args:
        n: A node to compute its activation tensor memory.
        node_nbits: A bit-width in which the node's weights should be quantized.

    Returns: The total memory of the node's activation tensor when quantized to the given bit-width.

    """

    origin_node = _get_origin_activation_node(n)
    node_output_size = origin_node.get_total_output_params()

    return node_output_size * node_nbits / BITS_TO_BYTES


class MpKpiMetric(Enum):
    """
    Defines kpi computation functions that can be used to compute KPI for a given target for a given mp config.
    The enum values can be used to call a function on a set of arguments.

     WEIGHTS_SIZE - applies the weights_size_kpi function

     ACTIVATION_OUTPUT_SIZE - applies the activation_output_size_kpi function

     TOTAL_WEIGHTS_ACTIVATION_SIZE - applies the total_weights_activation_kpi function

     BOPS_COUNT - applies the bops_kpi function

    """

    WEIGHTS_SIZE = partial(weights_size_kpi)
    ACTIVATION_OUTPUT_SIZE = partial(activation_output_size_kpi)
    TOTAL_WEIGHTS_ACTIVATION_SIZE = partial(total_weights_activation_kpi)
    BOPS_COUNT = partial(bops_kpi)

    def __call__(self, *args):
        return self.value(*args)



# Copyright 2021 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================


import os
from typing import List, Callable

import numpy as np
from PIL import Image

#:
FILETYPES = ['jpeg', 'jpg', 'bmp', 'png']


class FolderImageLoader(object):
    """

    Class for images loading, processing and retrieving.

    """

    def __init__(self,
                 folder: str,
                 preprocessing: List[Callable],
                 batch_size: int,
                 file_types: List[str] = FILETYPES):

        """ Initialize a FolderImageLoader object.

        Args:
            folder: Path of folder with images to load. The path has to exist, and has to contain at
            least one image.
            preprocessing: List of functions to use when processing the images before retrieving them.
            batch_size: Number of images to retrieve each sample.
            file_types: Files types to scan in the folder. Default list is :data:`~model_compression_toolkit.core.common.data_loader.FILETYPES`

        Examples:

            Instantiate a FolderImageLoader using a directory of images, that returns 10 images randomly each time it is sampled:

            >>> image_data_loader = FolderImageLoader('path/to/images/directory', preprocessing=[], batch_size=10)
            >>> images = image_data_loader.sample()

            To preprocess the images before retrieving them, a list of preprocessing methods can be passed:

            >>> image_data_loader = FolderImageLoader('path/to/images/directory', preprocessing=[lambda x: (x-127.5)/127.5], batch_size=10)

            For the FolderImageLoader to scan only specific files extensions, a list of extensions can be passed:

            >>> image_data_loader = FolderImageLoader('path/to/images/directory', preprocessing=[], batch_size=10, file_types=['png'])

        """

        self.folder = folder
        self.image_list = []
        print(f"Starting Scanning Disk: {self.folder}")
        for root, dirs, files in os.walk(self.folder):
            for file in files:
                file_type = file.split('.')[-1].lower()
                if file_type in file_types:
                    self.image_list.append(os.path.join(root, file))
        self.n_files = len(self.image_list)
        assert self.n_files > 0, f'Folder to load can not be empty.'
        print(f"Finished Disk Scanning: Found {self.n_files} files")
        self.preprocessing = preprocessing
        self.batch_size = batch_size

    def _sample(self):
        """
        Read batch_size random images from the image_list the FolderImageLoader holds.
        Process them using the preprocessing list that was passed at initialization, and
        prepare it for retrieving.
        """

        index = np.random.randint(0, self.n_files, self.batch_size)
        image_list = []
        for i in index:
            file = self.image_list[i]
            img = np.uint8(np.array(Image.open(file).convert('RGB')))
            for p in self.preprocessing:  # preprocess images
                img = p(img)
            image_list.append(img)
        self.next_batch_data = np.stack(image_list, axis=0)

    def sample(self):
        """

        Returns: A sample of batch_size images from the folder the FolderImageLoader scanned.

        """

        self._sample()
        data = self.next_batch_data  # get current data
        return data



# Copyright 2022 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

from typing import List

from model_compression_toolkit import KPI
from model_compression_toolkit.core.common.mixed_precision.mixed_precision_search_manager import \
    MixedPrecisionSearchManager
from model_compression_toolkit.core.common.quantization.candidate_node_quantization_config import \
    CandidateNodeQuantizationConfig
from model_compression_toolkit.core.common.logger import Logger
import numpy as np


def greedy_solution_refinement_procedure(mp_solution: List[int],
                                         search_manager: MixedPrecisionSearchManager,
                                         target_kpi: KPI) -> List[int]:
    """
    A greedy procedure to try and improve a mixed-precision solution that was found by a mixed-precision optimization
    algorithm.
    This procedure tries to increase the bit-width precision of configurable nodes that did not get the maximal
    candidate
    in the found solution.
    It iteratively goes over all such nodes, computes the KPI values on a modified configuration (with the node's next
    best candidate), filters out all configs that hold the KPI constraints and chooses one of them as an improvement
    step
    The choice is done in a greedy approach where we take the configuration that modifies the KPI the least.

    Args:
        mp_solution: A mixed-precision configuration that was found by a mixed-precision optimization algorithm.
        search_manager: A MixedPrecisionSearchManager object.
        target_kpi: The target KPIs for the mixed-precision search.

    Returns: A new, possibly updated, mixed-precision bit-width configuration.

    """
    # Refinement is not supported for BOPs KPI for now...
    if target_kpi.bops < np.inf:
        Logger.info(f'Target KPI constraint BOPs - Skipping MP greedy solution refinement')
        return mp_solution

    new_solution = mp_solution.copy()
    changed = True

    while changed:
        changed = False
        nodes_kpis = {}
        nodes_next_candidate = {}

        for node_idx in range(len(mp_solution)):
            if new_solution[node_idx] == 0:
                # layer has max config in the given solution, nothing to optimize
                continue

            node_candidates = search_manager.graph.get_configurable_sorted_nodes()[node_idx].candidates_quantization_cfg
            valid_candidates = _get_valid_candidates_indices(node_candidates, new_solution[node_idx])

            # Create a list of KPIs for the valid candidates.
            updated_kpis = []
            for valid_idx in valid_candidates:
                node_updated_kpis = search_manager.compute_kpi_for_config(
                    config=search_manager.replace_config_in_index(new_solution, node_idx, valid_idx))
                updated_kpis.append(node_updated_kpis)

            # filter out new configs that don't hold the KPI restrictions
            node_filtered_kpis = [(node_idx, kpis) for node_idx, kpis in zip(valid_candidates,updated_kpis) if
                               target_kpi.holds_constraints(kpis)]

            if len(node_filtered_kpis) > 0:
                sorted_by_kpi = sorted(node_filtered_kpis, key=lambda node_kpis: (node_kpis[1].total_memory,
                                                                               node_kpis[1].weights_memory,
                                                                               node_kpis[1].activation_memory))
                nodes_kpis[node_idx] = sorted_by_kpi[0][1]
                nodes_next_candidate[node_idx] = sorted_by_kpi[0][0]


        if len(nodes_kpis) > 0:
            # filter out new configs that don't hold the KPI restrictions
            node_filtered_kpis = [(node_idx, kpis) for node_idx, kpis in nodes_kpis.items()]
            sorted_by_kpi = sorted(node_filtered_kpis, key=lambda node_kpis: (node_kpis[1].total_memory,
                                                                           node_kpis[1].weights_memory,
                                                                           node_kpis[1].activation_memory))

            node_idx_to_upgrade = sorted_by_kpi[0][0]
            new_solution[node_idx_to_upgrade] = nodes_next_candidate[node_idx_to_upgrade]
            changed = True

    Logger.info(f'Greedy MP algorithm changed configuration from: {mp_solution} to {new_solution}')
    return new_solution


def _get_valid_candidates_indices(node_candidates: List[CandidateNodeQuantizationConfig],
                                  current_chosen_index: int) -> List[int]:
    """
    Find node's valid candidates to try and improve the node's MP chosen candidate.
    Valid indices are indices of candidates that have higher number of bits for both weights and activations.

    Args:
        node_candidates: Candidates of the node.
        current_chosen_index: Current index in MP configuration of the node.

    Returns:
        List of indices of valid candidates.
    """

    current_candidate = node_candidates[current_chosen_index]
    weights_num_bits = current_candidate.weights_quantization_cfg.weights_n_bits
    activation_num_bits = current_candidate.activation_quantization_cfg.activation_n_bits

    # Filter candidates that have higher bit-width for both weights and activations (except for the current index).
    return [i for i, c in enumerate(node_candidates) if c.activation_quantization_cfg.activation_n_bits >= activation_num_bits and c.weights_quantization_cfg.weights_n_bits >= weights_num_bits and not (c.activation_quantization_cfg.activation_n_bits == activation_num_bits and c.weights_quantization_cfg.weights_n_bits == weights_num_bits)]



# Copyright 2022 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from tensorflow.keras.layers import BatchNormalization, DepthwiseConv2D, Conv2DTranspose, Conv2D

from model_compression_toolkit.core.common.graph.graph_matchers import NodeOperationMatcher, \
    NodeFrameworkAttrMatcher
from model_compression_toolkit.core.common.substitutions.batchnorm_refusing import BatchNormalizationRefusing
from model_compression_toolkit.core.keras.constants import KERNEL, LINEAR, ACTIVATION, BIAS, GAMMA, \
    BETA, \
    MOVING_MEAN, MOVING_VARIANCE, EPSILON, USE_BIAS, LAYER_NAME
from model_compression_toolkit.core.keras.graph_substitutions.substitutions.batchnorm_folding import \
    update_kernel_for_bn_folding_fn


def batchnorm_refusing_node_matchers():
    """
    Function generates matchers for matching:
    (DepthwiseConv2D, Conv2D, Conv2DTranspose)[activation=linear] -> BatchNormalization.

    Returns:
        Matcher for batch norm nodes, and source nodes.
    """
    bn_node = NodeOperationMatcher(BatchNormalization)
    conv_node = NodeOperationMatcher(DepthwiseConv2D) | \
                NodeOperationMatcher(Conv2D) | \
                NodeOperationMatcher(Conv2DTranspose)

    activation_linear = NodeFrameworkAttrMatcher(ACTIVATION, LINEAR)
    source_node = conv_node & activation_linear
    return bn_node, source_node


def keras_batchnorm_refusing() -> BatchNormalizationRefusing:
    """
    Re-fuse BatchNormalization into preceding linear layers.

    Returns:
        A BatchNormalizationRefusing initialized for Keras models.
    """
    bn_node, source_node = batchnorm_refusing_node_matchers()
    return BatchNormalizationRefusing(source_node, bn_node, update_kernel_for_bn_folding_fn, KERNEL, BIAS, GAMMA, BETA,
                                      MOVING_MEAN, MOVING_VARIANCE, EPSILON, USE_BIAS, LAYER_NAME)



# Copyright 2022 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

from tensorflow_model_optimization.python.core.quantization.keras.quantize_config import QuantizeConfig
from typing import Tuple, List, Any, Dict
from tensorflow import Tensor
import six, abc


@six.add_metaclass(abc.ABCMeta)
class BaseQuantizeConfig(QuantizeConfig):
    """
    Base QuantizeConfig to define extra API methods needed by the GPTQ post-processing.
    """

    @abc.abstractmethod
    def get_quantization_variable(self):
        """
        A Functions that get the quantization parameters such as threshold, min, max ,etc.

        Returns: A list of trainable variable

        """

    @abc.abstractmethod
    def update_layer_quantization_params(self, layer) -> Tuple[Dict[str, Any],
                                                               Dict[str, Any],
                                                               Dict[str, Any]]:
        """
        A Function to calculate the needed change in attributes in NodeQuantizationConfig after retraining.
        Usually a function of the config quantizers.

        Args:
            layer: layer being quantized.

        Returns:
            3 dictionaries of attributes the quantize_config retraining has changed during GPTQ retraining.
            Keys must match NodeQuantizationConfig attributes:
            1. layer weights
            2. weight quantization config attributes
            3. activation quantization config attributes

        """

    @abc.abstractmethod
    def get_trainable_quantizer_parameters(self) -> List[Tensor]:
        """
        A function to get a list trainable of trainable parameters for GPTQ retraining from config quantizers

        Returns:
            A list of trainable Tensors

        """



# Copyright 2022 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

from typing import List, Any, Tuple

import torch

from model_compression_toolkit import FrameworkInfo
from model_compression_toolkit.core import common
from model_compression_toolkit.core.common import BaseNode
from model_compression_toolkit.core.common.graph.functional_node import FunctionalNode
from model_compression_toolkit.core.common.user_info import UserInformation
from model_compression_toolkit.core.pytorch.back2framework.instance_builder import node_builder
from model_compression_toolkit.core.pytorch.back2framework.pytorch_model_builder import PyTorchModelBuilder, \
    PytorchModel

from model_compression_toolkit.core.pytorch.default_framework_info import DEFAULT_PYTORCH_INFO
from model_compression_toolkit.core.pytorch.mixed_precision.mixed_precision_wrapper import PytorchMixedPrecisionWrapper


class MixedPrecisionPyTorchModel(PytorchModel):

    def __init__(self,
                 graph: common.Graph,
                 append2output=None,
                 fw_info: FrameworkInfo = DEFAULT_PYTORCH_INFO):
        """

        Args:
            graph: Graph to build its corresponding Pytorch model.
            append2output: List of nodes or OutTensor objects.
            fw_info: Framework information (e.g., mapping from layers to their attributes to quantize).
        """

        super().__init__(graph,
                         append2output,
                         fw_info)


    def _add_modules(self):
        configurable_nodes = self.graph.get_configurable_sorted_nodes()
        for n in self.node_sort:
            if n in configurable_nodes:
                self.add_module(n.name, PytorchMixedPrecisionWrapper(n, self.fw_info))
            else:
                if not isinstance(n, FunctionalNode):
                    self.add_module(n.name, node_builder(n))

    def _quantize_node_activations(self,
                                   node: BaseNode,
                                   input_tensors: List[torch.Tensor]) -> List[torch.Tensor]:
        """
        Quantize node's activation given input tensors.

        Args:
            node: Node to quantize its outputs.
            input_tensors: Input tensors of the node.

        Returns:
            Output of the node.

        """
        if node.is_all_activation_candidates_equal():
            # otherwise, we want to use the float tensor when building the model for MP search
            if isinstance(input_tensors, list):
                input_tensors = torch.cat(input_tensors, dim=0)
            input_tensors = node.candidates_quantization_cfg[0].activation_quantization_cfg.quantize_node_output(input_tensors)
        return input_tensors


    def _get_op_func(self,
                     node: BaseNode,
                     configurable_nodes_names: List[str]) -> Any:
        """
        Gets the operation function that runs the actual inference of the nodes compatible layer.

        Args:
            node: The corresponding node of the layer it runs.
            configurable_nodes_names: A list of names of configurable nodes in the quantized model.

        Returns: Module/functional to apply to the input tensors.

        """
        if node.name in configurable_nodes_names:
            return getattr(self, node.name)
        else:
            return node.type if isinstance(node, FunctionalNode) else getattr(self, node.name)




class MixedPrecisionPyTorchModelBuilder(PyTorchModelBuilder):
    """
    Mixed-precision PyTorch model.
    """
    def __init__(self,
                 graph: common.Graph,
                 append2output=None,
                 fw_info: FrameworkInfo = DEFAULT_PYTORCH_INFO,
                 return_float_outputs: bool = False):
        """

        Args:
            graph: Graph to build the model from.
            append2output: Nodes to append to model's output.
            fw_info: Information about the specific framework of the model that is built.
            return_float_outputs: Whether the model returns float tensors or not.
        """

        super().__init__(graph,
                         append2output,
                         fw_info,
                         return_float_outputs)

    def build_model(self) -> Tuple[PytorchModel, UserInformation]:
        """
        Build a PyTorch float model and return it.
        Returns: Float PyTorch model and user information.

        """
        return MixedPrecisionPyTorchModel(self.graph,
                                          self.append2output,
                                          self.fw_info), self.graph.user_info



# Copyright 2022 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import numpy as np
import tensorflow as tf
from keras.engine.base_layer import Layer
from tensorflow import TensorShape
from tensorflow_model_optimization.python.core.quantization.keras.quantizers import Quantizer

from model_compression_toolkit.core.common.quantization.quantizers.quantizers_helpers import uniform_quantize_tensor, \
    fix_range_to_include_zero
from model_compression_toolkit.core.common.target_platform import QuantizationMethod


class WeightsUniformQuantizer(Quantizer):
    """
    Quantizer for weights quantization.
    """

    def __init__(self,
                 nbits: int,
                 min_range: np.ndarray,
                 max_range: np.ndarray,
                 weight: np.ndarray,
                 quantization_method: QuantizationMethod):
        """

        Args:
            nbits: Number of bits to quantize.
            min_range: Min quantization range.
            max_range: Max quantization range.
            weight: Tensor of weights to quantize.
            quantization_method: Quantization method that is used (POT, Uniform, etc.)
        """

        super().__init__()

        min_range, max_range = fix_range_to_include_zero(np.array(min_range),
                                                         np.array(max_range),
                                                         nbits)
        self.weight = uniform_quantize_tensor(weight,
                                              range_min=min_range,
                                              range_max=max_range,
                                              n_bits=nbits).astype("float32")
        self.nbits = nbits
        self.min_range = min_range
        self.max_range = max_range
        self.delta = (self.max_range - self.min_range) / (2 ** self.nbits - 1)
        self.quantization_method = quantization_method

    def __call__(self, inputs, training, weights, **kwargs):
        """
        Apply quantization to the input tensor.

        Args:
            inputs: Input tensor to be quantized.
            training: Whether the graph is currently training.
            weights: Dictionary of weights the quantizer can use to quantize the tensor. This contains the weights created in the `build` function.
            **kwargs: Additional variables which may be passed to the quantizer.

        Returns:
            Quantized tensor.
        """
        with tf.name_scope('WeightsUniformQuant'):
            return self.weight

    def get_config(self):
        """

        Returns: Configuration of this WeightsUniformQuantizer

        """
        cfg = {"nbits": self.nbits,
               "min_range": self.min_range,
               "max_range": self.max_range,
               "weight": self.weight,
               "quantization_method": self.quantization_method
               }
        return cfg

    def build(self, tensor_shape: TensorShape, name: str, layer: Layer) -> dict:
        """
        Add variables under layer's scope.

        Args:
            tensor_shape: Shape of tensor which needs to be quantized.
            name: Name of tensor.
            layer: Layer to add variables to.

        Returns:
            Dictionary with new layer's variables.
        """
        return {}


# Copyright 2022 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from typing import Tuple, List

from model_compression_toolkit.core.common import BaseNode
from model_compression_toolkit.core.common.graph.memory_graph.cut import Cut
from model_compression_toolkit.core.common.graph.memory_graph.max_cut_astar import MaxCutAstar
from model_compression_toolkit.core.common.graph.memory_graph.memory_graph import MemoryGraph


def compute_graph_max_cut(memory_graph: MemoryGraph,
                          n_iter: int = 50,
                          astar_n_iter: int = 500,
                          eps: float = 1e-2) -> Tuple[List[BaseNode], float, List[Cut]]:
    """
    A wrapper function to compute max cut and schedule for a given model.
    It runs iterations of AStar search on the given memory graph with a dynamically updating estimation bound.

    Args:
        memory_graph: A MemoryGraph object to run the search on.
        n_iter: Limit on the number of AStar searches.
        astar_n_iter: Limit on the number of expansion iterations in a single AStar search.
        eps: Small value for defining a sufficient gap around an optimal solution in which the search would finish.

    Returns: A solution of the AStar search (schedule, max cut cost, cuts route).

    """
    max_cut_astar = MaxCutAstar(memory_graph=memory_graph)
    last_result = (None, 0, None)
    l_bound = memory_graph.memory_lbound_single_op
    u_bound = 2 * sum([t.total_size for t in memory_graph.b_nodes]) - l_bound
    it = 0
    while it < n_iter:
        estimate = (u_bound + l_bound) / 2
        schedule, max_cut_size, cuts = max_cut_astar.solve(estimate_factor=estimate, iter_limit=astar_n_iter)
        if schedule is None:
            return last_result

        next_u_bound = min(estimate, max_cut_size)
        last_result = (schedule, max_cut_size, cuts)

        if l_bound * (1 + eps) >= next_u_bound:
            return last_result

        it += 1

    return last_result
